from pathlib import Path

import pytest

from swap_env.dotenv_dir import DOTENV_DIR_NAME, iter_tree, search_fs_for_dotenv_dir


@pytest.fixture
def tmp_home(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Path:
    monkeypatch.setattr(Path, "home", lambda: tmp_path)
    return tmp_path


class TestSearchFSForDotenvDir:
    def test_dotenv_dir_in_cwd_found(
        self, tmp_home: Path, monkeypatch: pytest.MonkeyPatch
    ):
        (cwd := tmp_home.joinpath("current_dir")).mkdir()
        (dotenv_dir := cwd.joinpath(DOTENV_DIR_NAME)).mkdir()
        monkeypatch.chdir(cwd)

        assert search_fs_for_dotenv_dir() == dotenv_dir

    def test_dotenv_dir_in_intermediate_dir_found(
        self, tmp_home: Path, monkeypatch: pytest.MonkeyPatch
    ):
        (mid_dir_one := tmp_home.joinpath("mid_dir_one")).mkdir()
        (mid_dir_two := mid_dir_one.joinpath("mid_dir_two")).mkdir()
        (cwd := mid_dir_two.joinpath("current_dir")).mkdir()
        (dotenv_dir := mid_dir_one.joinpath(DOTENV_DIR_NAME)).mkdir()
        monkeypatch.chdir(cwd)

        assert search_fs_for_dotenv_dir() == dotenv_dir

    def test_default_to_home_if_none_found(
        self, tmp_home: Path, monkeypatch: pytest.MonkeyPatch
    ):
        (mid_dir := tmp_home.joinpath("mid_dir_one")).mkdir()
        (cwd := mid_dir.joinpath("current_dir")).mkdir()
        monkeypatch.chdir(cwd)

        assert search_fs_for_dotenv_dir() == tmp_home.joinpath(DOTENV_DIR_NAME)

    def test_default_to_home_if_cwd_not_in_home(self, tmp_home: Path, tmp_cwd: Path):
        tmp_cwd.joinpath(DOTENV_DIR_NAME).mkdir()

        assert search_fs_for_dotenv_dir() == tmp_home.joinpath(DOTENV_DIR_NAME)


class TestIterTree:
    def test_empty_if_start_dir_not_in_end_dir(self, tmp_path: Path):
        (start_dir := tmp_path.joinpath("a")).mkdir()
        (end_dir := tmp_path.joinpath("b")).mkdir()

        assert list(iter_tree(start_dir, end_dir)) == []

    def test_empty_if_end_dir_in_start_dir(self, tmp_path: Path):
        (start_dir := tmp_path.joinpath("a")).mkdir()
        (end_dir := start_dir.joinpath("b")).mkdir()

        assert list(iter_tree(start_dir, end_dir)) == []

    def test_start_dir_returned_if_directly_in_end_dir(self, tmp_path: Path):
        (end_dir := tmp_path.joinpath("a")).mkdir()
        (start_dir := end_dir.joinpath("b")).mkdir()

        assert list(iter_tree(start_dir, end_dir)) == [start_dir]

    def test_correct_dirs_returned_when_nested(self, tmp_path: Path):
        (end_dir := tmp_path.joinpath("a")).mkdir()
        (third_dir := end_dir.joinpath("b")).mkdir()
        (second_dir := third_dir.joinpath("c")).mkdir()
        (start_dir := second_dir.joinpath("d")).mkdir()

        third_dir.joinpath("irrelevant").mkdir()

        assert list(iter_tree(start_dir, end_dir)) == [start_dir, second_dir, third_dir]



from pathlib import Path

import pytest

from swap_env.dotenv_files import DotenvFiles


class TestDotenvFiles:
    def test_value_error_raised_given_a_file(self, tmp_path: Path):
        (filepath := tmp_path.joinpath("file")).touch()

        with pytest.raises(ValueError, match="Path must be a directory"):
            DotenvFiles(path=filepath)

    def test_dir_created_if_not_exists(self, tmp_path: Path):
        path = tmp_path.joinpath("dir")

        DotenvFiles(path=path)
        assert path.is_dir()

    def test_dotenv_files_loaded_with_suffix_names(self, tmp_path: Path):
        (file_one := tmp_path.joinpath(".env.one")).touch()
        (file_two := tmp_path.joinpath(".env.two")).touch()

        dotenv_files = DotenvFiles(path=tmp_path)
        assert dotenv_files._files == {"one": file_one, "two": file_two}

    def test_files_without_dotenv_prefix_ignored(self, tmp_path: Path):
        (file_one := tmp_path.joinpath(".env.one")).touch()
        tmp_path.joinpath(".two").touch()

        dotenv_files = DotenvFiles(path=tmp_path)
        assert dotenv_files._files == {"one": file_one}

    def test_link_created_in_cwd(self, tmp_path: Path, tmp_cwd: Path):
        (dotenv := tmp_path.joinpath(".env.one")).touch()
        dotenv_files = DotenvFiles(tmp_path)
        dotenv_files.link("one")

        dotenv_link = tmp_cwd.joinpath(".env")
        assert dotenv_link.samefile(dotenv)

    def test_link_overwritten_if_exists(self, tmp_path: Path, tmp_cwd: Path):
        (dotenv := tmp_path.joinpath(".env.one")).touch()
        (initial_link := tmp_path.joinpath("initial-link")).touch()

        dotenv_link = tmp_cwd.joinpath(".env")
        dotenv_link.symlink_to(initial_link)

        dotenv_files = DotenvFiles(tmp_path)
        dotenv_files.link("one")

        assert dotenv_link.samefile(dotenv)

    def test_file_overwritten_if_exists(self, tmp_path: Path, tmp_cwd: Path):
        (dotenv := tmp_path.joinpath(".env.one")).touch()
        (existing_dotenv := tmp_cwd.joinpath(".env")).touch()

        dotenv_files = DotenvFiles(tmp_path)
        dotenv_files.link("one")

        assert existing_dotenv.samefile(dotenv)

    def test_local_dotenv_file_saved(self, tmp_path: Path, tmp_cwd: Path):
        tmp_cwd.joinpath(".env").write_text("#dotenv")

        dotenv_files = DotenvFiles(tmp_path)
        dotenv_files.save_local_dotenv("name")

        assert tmp_path.joinpath(".env.name").read_text() == "#dotenv"



from pathlib import Path

from prompt_toolkit.shortcuts import choice, confirm, prompt
from prompt_toolkit.validation import Validator

from swap_env.dotenv_dir import search_fs_for_dotenv_dir
from swap_env.dotenv_files import DotenvFiles


def prompt_to_save_current_file_if_exists(dotenv_files: DotenvFiles) -> None:
    local_dotenv_path = Path.cwd().joinpath(".env")
    if not local_dotenv_path.exists() or local_dotenv_path.is_symlink():
        return

    save_local_file = confirm(
        "A .env file already exists, do you want to save this before continuing?"
    )

    if not save_local_file:
        return

    new_dotenv_name = prompt(
        "Enter a name for this .env file: ",
        validator=Validator.from_callable(
            lambda name: name not in dotenv_files and name != "",
            error_message="The name is empty or already exists",
        ),
    )

    dotenv_files.save_local_dotenv(name=new_dotenv_name)


def app():
    dotenv_dir = search_fs_for_dotenv_dir()
    dotenv_files = DotenvFiles(path=dotenv_dir)

    prompt_to_save_current_file_if_exists(dotenv_files=dotenv_files)

    if not dotenv_files:
        print(f"No dotenv files found in {dotenv_files.path}")
        return

    dotenv_file_choice = choice(
        "Select a .env file:",
        options=[(name, name) for name in dotenv_files],
    )

    dotenv_files.link(dotenv_file_choice)



from collections.abc import Generator
from pathlib import Path

DOTENV_DIR_NAME = ".swap-env"


def search_fs_for_dotenv_dir() -> Path:
    """Search for a dotenv dir from the CWD up to the home dir.

    If no dotenv dir is found, or if the CWD isn't inside the home dir, then
    this defaults to `~/.swap-env/`.
    """
    cwd = Path.cwd()
    home = Path.home()

    return next(
        (
            dotenv_dir
            for dir in iter_tree(cwd, home)
            if (dotenv_dir := dir.joinpath(DOTENV_DIR_NAME)).is_dir()
        ),
        home.joinpath(DOTENV_DIR_NAME),
    )


def iter_tree(start_dir: Path, end_dir: Path) -> Generator[Path, None, None]:
    """Return an iterator of paths from start_dir up to end_dir.

    If start_dir is not in end_dir, an empty iterator is returned.
    """
    if not start_dir.is_relative_to(end_dir):
        return

    current_dir = start_dir
    while current_dir != end_dir:
        yield current_dir
        current_dir = current_dir.parent



from collections.abc import Iterator
from pathlib import Path


class DotenvFiles:
    """Collection of .env files named `.env.<name>` to be linked to from the cwd.

    On initialisation, a directory is created at the given path if it doesn't
    exist. All files in the given directory named `.env.*` are picked up and can
    be accessed by their name, given by the part after `.env.`.

    Example usage::

        >>> dotenv_files = DotenvFiles(pathlib.Path.home())

        >>> dotenv_files.save_local_dotenv("test")

        >>> list(dotenv_files)
        ['dev', 'test']

        >>> # Link ./.env to ~/.env.dev
        >>> dotenv_files.link("dev")
    """

    DOTENV_FILE_PREFIX = ".env."

    def __init__(self, path: Path):
        if not path.exists():
            path.mkdir()
        if not path.is_dir():
            raise ValueError("Path must be a directory")

        self.path = path
        self._local_dotenv = Path.cwd().joinpath(".env")
        self._load()

    def _load(self) -> None:
        """Load .env files with names given by their suffix."""
        self._files = {
            file.name.removeprefix(self.DOTENV_FILE_PREFIX): file
            for file in self.path.iterdir()
            if file.name.startswith(self.DOTENV_FILE_PREFIX)
        }

    def __iter__(self) -> Iterator[str]:
        """Return an alphabetically sorted iterator of .env file nicknames."""
        return iter(sorted(self._files))

    def __getitem__(self, name: str) -> Path:
        """Return the path of the .env file with the nickname `name`."""
        return self._files[name]

    def __bool__(self) -> bool:
        """Return `True` iff there are files named `.env.*` in the given dir."""
        return bool(self._files)

    def link(self, name: str) -> None:
        """Create symlink from ./.env to the dotenv file with the given name."""
        self._local_dotenv.unlink(missing_ok=True)
        self._local_dotenv.symlink_to(self._files[name])

    def save_local_dotenv(self, name: str) -> None:
        """Save ./.env to the store with the given name."""
        new_path = self._local_dotenv.rename(
            self.path.joinpath(f"{self.DOTENV_FILE_PREFIX}{name}")
        )
        self._files[name] = new_path
